(* Copyright 2010 Sarah Winkler
 * GNU Lesser General Public License
 *
 * This file is part of MKBtt.
 * 
 * MKBtt is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * MKBtt is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with MKBtt. If not, see <http://www.gnu.org/licenses/>.
 *)

(** Main control for (o)MKBtt
 @author Sarah Winkler
 @since  2010/11/15 *)

(* This module contains the main control loop for (ordered or standard) 
   multicompletion, and some auxiliary functions.  *)

(*** OPENS (1) ***********************************************************)
open Util;;

(*** SUBMODULES **********************************************************)
module C = Completion;;
module W = World;;
module Monad = W.Monad;;
module St = Statistics;;
module CP = CompletionProcessx;;
module N = IndexedNode;;
module NS = NodeState;;
module NSA = NodeSetAux;;
module NI = NodeTermIndexx;;
module PIso = ProcessIsomorphism;;
module I = Inferences;;
module G = GoalWrapper;;
module SS = SelectionStrategy;;

module Option = Util.Option;;

(*** OPENS (2) ***********************************************************)
(*open Monad;;*)
let (>>=) = Monad.(>>=);;
let (>>) = Monad.(>>);;
let return = Monad.return;;
let map = Monad.map;;

(*** TYPES ***************************************************************)
type success_type = Proof | Saturation

(*** FUNCTIONS ***********************************************************)

let log1 s = W.log 1 s

let log2 s = (*Format.printf "%s\n" s; *)W.log 2 s

let log_iteration i = 
 St.inc_n_iterations  >>
 log1 ("Starting iteration "^(string_of_int i)) >> 
 NS.open_nodes >>= IndexedSet.to_stringm >>= fun  os ->
 NS.closed_nodes >>= IndexedSet.to_stringm >>= fun  cs ->
 log1 ("Open nodes: \n"^os^"\nClosed nodes:\n"^cs)
;;

let log_choose n = 
 N.to_stringm n >>= fun ns -> 
 log1 ("CHOOSE "^ns)
;;

let log_orient n = N.to_stringm n >>= fun ns -> log2 (ns^" oriented")

let log_deduce ns =
 IndexedSet.to_stringm ns >>= fun s -> 
 log2 ("After deduce & rewrite: "^s) (*>>
 NI.log_index_state*)
;;

let union = List.union

(* ------------------------------------------------------------------ *)
(*  CHECK SUCCESS                                                     *)
(* ------------------------------------------------------------------ *)

let szs_proof = (*function
 | Proof -> ( *)
  W.get_options >>= fun o -> match Option.the (C.conjecture o) with
  | C.Conjecture C.Equality -> failwith "unknown SZS status"
  | C.Conjecture C.Disequality -> failwith "unknown SZS status"
  | C.NegatedConjecture C.Equality -> return C.Uns (* COL059-1 *)
  | C.NegatedConjecture C.Disequality -> return C.Uns
  (* )
 | Saturation -> (
  St.options >>= fun o -> match C.conjecture o with
  | C.Conjecture C.Equality -> failwith "unknown SZS status"
  | C.Conjecture C.Disequality -> failwith "unknown SZS status"
  | C.NegatedConjecture C.Equality -> return C.Noc
  | C.NegatedConjecture C.Disequality -> return C.Sat
  )*)
;;

let completion_finished i =
 (*log2 "check completion finished" >>*)
 I.unfinished_processes >>= fun uc ->
 CP.complement uc >>= function
 | [] -> 
  NS.non_empty_open_nodes_exist >>= fun b ->
  return (if b then None else Some (C.Gup "Completion failed.", None))
 | p :: _ ->
  (*log2 ("successful process "^(CP.to_string p)) >>*)
  (*(if verbose () > 0 then print_winner p n_c);*)
  return (Some (C.Sat, Some (C.MKBttResult p)))
;;

let proof_finished i = 
 G.proven_for >>= function
  | [] ->
   I.unfinished_processes >>= fun uc ->
   CP.complement uc >>= ( function 
    | [] ->
     return None
    | p :: _ ->
     G.decide p >>= fun res ->
     return (Some(res,Some (C.MKBttResult p))))
  | p :: _ ->
   szs_proof >>= fun szs ->
   return (Some (szs, Some (C.MKBttResult p)))
;;

let success i =
 W.get_options >>= fun o ->
 match C.mode o with
  | C.Completion -> completion_finished i 
  | _ -> proof_finished i 
;;

(* ------------------------------------------------------------------ *)
(*  CHOICE CONTROL                                                    *)
(* ------------------------------------------------------------------ *)

let choose i =
 W.get_options >>= fun o ->
 let tstart = Unix.gettimeofday() in
 let sm = C.selection_macro o in
 let choose = 
  (*if sm = Some(SS.Old) then
   Choose.best_m
  else*) if sm = Some(SS.JapanSum) || sm = Some(SS.JapanMax) then
   ChooseHaruhiko.bestl ()
  else
   SelectionStrategyAux.apply (C.selection_strategy o)
 in
 choose >>= fun result ->
 St.add_t_selection (Unix.gettimeofday () -. tstart) >>
 return result
;;

(* ------------------------------------------------------------------ *)
(*  ISOMORPHISM CHECKS                                                *)
(* ------------------------------------------------------------------ *)
let remove_isomorphic i =
 NS.all_nodes >>= PIso.filter_subsumed_sometimes >>= NS.remove_processes
;;


(* ------------------------------------------------------------------ *)
(*  CONTROL LOOP                                                      *)
(* ------------------------------------------------------------------ *)

let id = return ()

let rewrite_open_with_new_node singleindex =
 NS.open_nodes >>= fun no -> 
 MKBtt.rewrite_open_with no singleindex >>= fun no' ->
 MKBtt.rewrite_open no' >>= fun no'' ->
 NS.add_open (union no' no'') >>
 NS.open_nodes >>= MKBtt.gc >>= NS.set_open
;;


let rec iterate i =
 log_iteration i >>
 success i >>= function
  | Some r -> return r
  | None ->
 W.get_options >>= fun o ->
 (if C.check_isomorphisms_repeatedly o then remove_isomorphic i else id) >>
 choose i >>= fun n -> NS.remove_open n >>
 (*log_choose n >>*)
 MKBtt.rewrite_open [n] >>= NS.add_open >>
 N.has_non_empty_labels n >>= fun not_superfluous ->
 if not_superfluous then
  MKBtt.orient n >> 
  log_orient n >>
  N.contains_rule_labels n >>= fun is_orientable ->
  if is_orientable || (C.is_ordered o) then
  (*log2 "REWRITE CLOSED" >>*)
   NS.single_rindex n >>= fun idx ->
   NS.closed_nodes >>= fun nc -> 
   MKBtt.rewrite_closed_with idx nc >>= fun no' ->
   MKBtt.rewrite_open no' >>= fun no'' ->
   MKBtt.gc (union no' no'') >>= NS.add_open >> MKBtt.gc nc >> (* ! *)
   G.rewrite n idx >>
   NS.index_add_closed_node n >>
   (*log2 "DEDUCE" >>*)
   I.deduce_rewrite n >>= MKBtt.gc >>= fun ns -> 
   (*log_deduce ns >>*)
   NS.add_open ns >>
   (* filter nopen here? yes - e.g. 3.30 much faster. *)
   (*log2 "REWRITE OPEN" >>*)
   rewrite_open_with_new_node idx >>
   (* *)
   iterate (i+1)
  else
   NS.index_add_closed_node n >> iterate (i+1)
 else iterate (i+1)
;;

(* ------------------------------------------------------------------ *)
(*  STARTING COMPLETION                                               *)
(* ------------------------------------------------------------------ *)

let complete eqs goal = 
 G.set goal >>
 NSA.of_axioms eqs >>= NS.add_open >>
 iterate 0
;;

let completion_output p ns =
 History.joint_history p [] ns >>= fun h ->
 W.get_options >>= fun o ->
 History.as_string p (C.filename o) h >>= fun s ->
 return (s, "Ground complete system derived.")
;;

let print_proof s p =
 let prooftype = if s=C.Uns then "Refutation" else "Proof" in
 let rec l i = if i<=0 then "" else "-"^(l (i-1)) in
 let line () = Format.printf "%s%s%s\n" "%" (l 76) "%" in
 W.get_options >>= fun o ->
 (NS.er_contain_closed p >>= fun ns ->
  if C.has_goal o then G.proof_output s p ns
  else completion_output p ns) >>= fun (proof, explanation) ->
 line (); 
 Format.printf "oMKBtt result: %s\n" explanation;
 line ();
 let f = C.filename o in
 Format.printf "%s SZS output start %s for %s\n" "%" prooftype f;
 Format.printf "%s" proof;
 Format.printf "%s SZS output end\n" "%";
 line ();
 return ()
;;

