(* Copyright 2010 Sarah Winkler
 * GNU Lesser General Public License
 *
 * This file is part of MKBtt.
 * 
 * MKBtt is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * MKBtt is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with MKBtt. If not, see <http://www.gnu.org/licenses/>.
 *)

(** Switch isomorphism checks
 @author Sarah Winkler
 @since  2009/08/11
*)

(*** OPENS ***************************************************************)
open Util;;

(*** SUBMODULES **********************************************************)
module Rewriting = Processors.Rewritingx;;
module Trs = U.Trs;;
module C = Completion;;
module NS = IndexedSet;;
module CP = CompletionProcessx;;
module R = TrsRenaming;;
module P = TermPermutation;;
module W = World;;
module Monad = W.Monad;;

(*** OPENS ***************************************************************)
open Monad;;
open World;;

(*** FUNCTIONS ***********************************************************)
let symmetric_for_renaming rule trs p nodeset =
 NodeSetAux.project_e nodeset p >>= fun e_p ->
 NodeSetAux.project_r nodeset p >>= fun r_p ->
 try
  let theta = R.orientation_symmetric rule trs in
  return ((R.eqns_invariant e_p theta) && (R.trs_invariant r_p theta))
 with
  | R.Not_renamable -> return false
;;

let symmetric_for_permutation rule trs p nodeset =
 try
  let pi = P.orientation_symmetric rule trs in
  NodeSetAux.project_e nodeset p >>= fun e_p ->
  NodeSetAux.project_r nodeset p >>= fun r_p ->
  let is_symmetric = 
   (P.equations_invariant e_p pi) && (P.trs_invariant r_p pi) 
  in
(*  Format.printf "Rule %s is symmetric for \nEquations:%s\n Rules:%s\n Constraints:%s\n" (Rule.to_string rule) (Trs.to_string e_p) (Trs.to_string r_p) (Trs.to_string c_p);*)
  return is_symmetric
 with
  | P.Not_permutable -> return false
;;

let orientation_symmetric rule trs p nodeset options =
 let answer = 
 match C.check_isomorphism options with
  | C.NoChecks -> return false
  | C.Renamings -> symmetric_for_renaming rule trs p nodeset
  | C.Permutations -> symmetric_for_permutation rule trs p nodeset
 in
 answer
;;

let print_iso (e, r, c) (e', r', c') =
 Format.printf "The following two processes are isomorphic:\n";
 Format.printf "Equations1:%s\n Rules1:%s\n Constraints1:%s\n"
  (Equation.all_to_string e) (Trs.to_string r) (Trs.to_string c);
 Format.printf "Equations1:%s\n Rules1:%s\n Constraints1:%s\n"
  (Equation.all_to_string e') (Trs.to_string r') (Trs.to_string c');
;;

(* checks whether an isomorphism between these two processes exists *)
let exists p p' =
 W.get_options >>= function o ->
 match C.check_isomorphism o with
  | C.NoChecks -> return false
  | C.Renamings -> return (R.processes_match p p')
  | C.Permutations -> return (P.processes_match p p')
;;

let get_projections nodeset p = 
 NodeSetAux.project_e nodeset p >>= fun e_p ->
 NodeSetAux.project_r nodeset p >>= fun r_p ->
 NodeSetAux.project_c nodeset p >>= fun c_p ->
 return (p, (e_p, r_p, c_p))
;;

let print_process_state nodeset p =
 get_projections nodeset p >>= fun (_, (e, r, c)) ->
 Format.printf "%s has Equations:%s\n Rules1:%s\n Constraints1:%s\n"
  (CP.to_string p)
  (Equation.all_to_string e) (Trs.to_string r) (Trs.to_string c);
 return ()
;;

(* returns all processes that are subsumed by others.
   n(n-1)/2 exist-checks *)
let filter_subsumed ps nodeset =
 let rec filter_subsumed' = function
   [] -> return []
  | p :: ps ->
   filter (fun p' -> exists p p') ps >>= fun subsumed ->
   let ps' = List.diff ps subsumed in
   filter_subsumed' ps' >>= fun ps'' ->
   return (List.union (List.map fst subsumed) ps'')
 in
 map (get_projections nodeset) ps >>= fun states ->
 filter_subsumed' states
;;

let every_second_time () = let r = Random.float 1.0 in (r < 0.5)

(* perform periodical isomorphism checks only sometimes, currently 
   every second iteration *)
let filter_subsumed_sometimes ns =
 CompletionProcessx.all_processes >>= fun pset ->
 if every_second_time () then 
  filter_subsumed pset ns 
 else 
  return []
;;
