(* Copyright 2010 Sarah Winkler
 * GNU Lesser General Public License
 *
 * This file is part of MKBtt.
 * 
 * MKBtt is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * MKBtt is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with MKBtt. If not, see <http://www.gnu.org/licenses/>.
 *)

(** Check for renaming isomorphisms
 @author Sarah Winkler
 @since  2009/07/07 *)

(*** OPENS ********************************************************************)
open Util;;
(*** SUBMODULES **********************************************************)
module Var = Rewriting.Variable;;
module Fun = Rewriting.Function;;
module Term = U.Term;;
module Rule = U.Rule;;
module Trs = U.Trs;;
module Sub = Termx.VSub;;
module CP = CompletionProcessx;;
module Tx = Termx;;

module SymbolSubstitution = struct
 include Map.Make (Fun)(Fun);;

 exception Not_injective;;

 let replace x t sub = add x t sub;;

 let union sub sub' = fold (fun q t sub -> add q t sub) sub' sub;;
 
 (* singleton substitution mapping f to g *)
 let singleton f g = add f g empty;;

 (* apply a substitution to some other substitution *)
 let apply symsub substitute symsub' =
  fold
   (fun f g s -> replace f (substitute symsub g) s)
   symsub'
   empty
 ;;

 let is_injective t =
  let check _ f = function 
    None -> None 
   | Some d -> if List.mem f d then None else Some (f :: d)
  in
  match fold check t (Some []) with None -> false | _ -> true
 ;;
(*
 let is_injective t =
 try
  ignore
   (fold
    (fun _ t d ->
     if List.mem t d then raise Not_injective else t :: d
    )
    t
    []
   );
  true
 with
  | Not_injective -> false
 ;;
*)
 let to_string s =
  fold
  (fun key data str -> 
    str ^ (Fun.to_string key) ^ ":" ^ (Fun.to_string data) ^ " ")
  s
  ""
 ;;

 let rec rename_term theta = function
  | Term.Var x as t -> t
  | Term.Fun (f, ts) -> 
   begin
    let f' = try find f theta with Not_found -> f in
    let ts' = List.map (rename_term theta) ts in
    Term.Fun (f', ts')
   end
 ;;

 let rename_rule theta = 
  (uncurry Rule.of_terms) <.> (Rule.map (rename_term theta))
 ;;

 let rename_trs trs theta = Trs.of_list (Trs.map (rename_rule theta) trs)

 let rename_equation theta e = Equation.map (rename_term theta) 

 let rename_eqs eqs theta = List.map (rename_equation theta) eqs
 end

module SymSub = SymbolSubstitution;;

(*** EXCEPTIONS **********************************************************)
exception Substitution_inconsistent;;
exception Not_renamable;;

(*** FUNCTIONS ***********************************************************)

let map_pair f (l, r) = (f l, f r);;

(* check whether two terms are variants with respect to a variable
   renaming mgu and a symbol renaming theta. mgu and theta are extended
   if consistent and returned as a pair *)
 let rec variants mgu theta = function
  | [] -> (mgu, theta)
  | ((Term.Var y), (Term.Var x)) :: eqs ->
    begin
    try
     let z = Sub.find y mgu in
     if (Var.compare z x) == 0 then
      variants mgu theta eqs
     else raise Not_renamable
    with Not_found ->
     (
     let s = Sub.singleton y x in
     variants (Sub.union mgu s) theta eqs
     )
     end
  | (Term.Fun (f0, ts0), Term.Fun (f1, ts1)) :: eqs  ->
   if (List.length ts0) == (List.length ts1) then
    begin
    try
     let g = SymSub.find f0 theta in 
     if (Fun.compare g f1) == 0 then
      let eqs' = (List.combine ts0 ts1) @ eqs in
      variants mgu theta eqs'
     else raise Not_renamable
    with Not_found ->
     (
     let s = SymSub.singleton f0 f1 in
     let eqs' = (List.combine ts0 ts1) @ eqs in
     variants mgu (SymSub.union s theta) eqs')
    end
   else
    raise Not_renamable
  | _ -> raise Not_renamable
;;

let term_variants s t mgu theta = 
 let (sigma, theta) = variants mgu theta [(s, t)] in
 if (SymSub.is_injective theta) && (Sub.is_bijective sigma) then
  (sigma, theta)
 else
  raise Not_renamable
;;

let terms_are_variants s t = 
 try
  let sub = Sub.empty in
  let symsub = SymSub.empty in
  let _ = term_variants s t sub symsub in
  true
 with
  | Not_renamable -> false
;;

let rule_variants rl1 rl2 theta =
  let sigma = Sub.empty in (* variable renaming *)
  let l1, r1 = Rule.lhs rl1, Rule.rhs rl1 in (* check lhs *)
  let l2, r2 = Rule.lhs rl2, Rule.rhs rl2 in (* check rhs *)
  let sigma', theta' = term_variants l1 l2 sigma theta in
  snd (term_variants r1 r2 sigma' theta') (* return theta'' *)
;;

let rules_are_variants r1 r2 =
 try
  let theta = SymSub.empty in
  let _ = rule_variants r1 r2 theta in
  true
 with
  | Not_renamable -> false
;;

(* returns a symbol renaming if a rule is isomorphic to its invert,
   otherwise raises Not_renamable *)
let rule_symmetry rule theta = rule_variants rule (Rule.invert rule) theta

let rule_symmetric rule = 
 try
  let theta = SymSub.empty in 
  let t = rule_symmetry rule theta in ignore t; true
 with
  | Not_renamable -> false
;;

let orientation_symmetric rule trs =
  let theta = SymSub.empty in
  (* try to find renaming matching the rule with its invert *)
  let theta = rule_variants rule (Rule.invert rule) theta in
  (* apply renaming to trs *)
  let trs' = SymSub.rename_trs trs theta in
  (* check if resulting trs is equal to original one - in this case
     the two rule orientations are redendant *)
  if Trs.equivalent trs trs' then theta else 
  (
  Trs.fprintf Format.std_formatter trs;
  Trs.fprintf Format.std_formatter trs';
  raise Not_renamable)
;;

(*********** check equality of equation sets ****************************)
let reverse rule = Rule.of_terms (Rule.rhs rule) (Rule.lhs rule)

let remove_eqn rule trs =
  if Trs.mem rule trs then
   Trs.remove rule trs
  else if Trs.mem (reverse rule) trs then
   Trs.remove (reverse rule) trs
  else raise Not_found
;;

(*
(* eqn is actually trs *)
let equations_invariant eqn theta =
 let eqnren = SymSub.rename_eqs eqn theta in
 equal_as_equations eqn eqnren
;;

let exists_rule_variant trs rule =
 Trs.exists (Rule.is_variant rule) trs
;;

let equal_as_trs trs1 trs2 =
  Trs.for_all (exists_rule_variant trs1) trs2
;;
*)

let trs_invariant trs theta =
 let trsren = SymSub.rename_trs trs theta in
 Trs.equivalent trs trsren
;;

let eqns_invariant eqns =
 let rs = List.map ((uncurry Rule.of_terms) <.> Equation.terms) eqns in
 let trs = Trs.of_list rs in
 trs_invariant trs
;;

(********* functions for checking arbitrary processes equal *************)

(* return option on possible extension of symbol substitution theta that
   makes eqn isomorphic to eqn' (checks only one direction)             *)
let possible_theta rule rule' theta =
 try let theta' = rule_variants rule rule' theta in [theta']
 with Not_renamable -> []
;;

(* abstract from equations vs. rules - the followiung works for both,
   provided that theta_set_for_termpair is set to respective function
   *)

(* returns set of possible extensions of theta' of theta such that e is
   isomorphic to e' under theta'. Assumes |e|=|e'| > 0                  *)
let rec match_termpair_sets_for theta_set_for_termpair e e' theta =
 let eqn = Trs.choose e in
 let add eqn' tset = (theta_set_for_termpair eqn eqn' theta) :: tset in
 let theta_e_set = Trs.fold add [] e' in
 let match_eqn (eqn', thetaset') =
  let e_red, e_red' = Trs.remove eqn e, Trs.remove eqn' e' in
  match_termpair_sets theta_set_for_termpair e_red e_red' thetaset'
 in
 List.concat (List.map match_eqn theta_e_set)
(* theta_set is a set of function symbol renamings. This function filters
   all theta \in theta_set which can be extended to theta' such that e and
   e' are isomorphic under theta'.
   Assumes |e|=|e'|.
   Returns set of all possibilities for theta'.                         *)
and match_termpair_sets theta_set_for_termpair e e' theta_set =
 if (Trs.is_empty e) then
  theta_set
 else
  let f = theta_set_for_termpair in
  List.concat (List.map (match_termpair_sets_for f e e') theta_set)
;;

(* and the same again for equations *)
let rec match_equations_sets_for theta_set_for_eqns e e' theta =
 let eqn = List.hd e in
 let add tset eqn' = (theta_set_for_eqns eqn eqn' theta) :: tset in
 let theta_e_set = List.fold_left add [] e' in
 let match_eqn (eqn', thetaset') =
  let e_red, e_red' = List.remove eqn e, List.remove eqn' e' in
  match_equations_sets theta_set_for_eqns e_red e_red' thetaset'
 in
 List.concat (List.map match_eqn theta_e_set)

and match_equations_sets theta_set_for_eqns e e' theta_set =
 if e = [] then theta_set
 else
  let f = theta_set_for_eqns in
  List.concat (List.map (match_equations_sets_for f e e') theta_set)
;;

(********** match equation sets *****************************************)

(* checks if two equations are isomorphic under theta, basically 
   achieved by the function f. 
   For TrsRenaming, this is possible_theta                              *)
let theta_set_for_eqn_pair f eqn eqn' theta =
 let rl = Equation.to_rule eqn in
 let rl' = Equation.to_rule eqn' in
 let rlr = f rl rl' theta in
 let rrl = f (reverse rl) rl' theta in
 (eqn', rlr @ rrl)
;;

(* theta_set is a set of function symbol renamings. This function filters
   all theta \in theta_set which can be extended to theta' such that e and
   e' are isomorphic under theta'.
   Assumes |e|=|e'|.
   Returns set of all possibilities for theta'.                         *)
let match_equation_sets =
 match_equations_sets (theta_set_for_eqn_pair possible_theta)
;;

(********** match rule sets *********************************************)

(* checks if two rules are isomorphic under theta, basically achieved by
   the function f. For TrsRenaming, this is possible_theta *)
let theta_set_for_rule_pair f rule rule' theta =
 let thetaset =  f rule rule' theta in
  (rule', thetaset)
;;

(* theta_set is a set of function symbol renamings. This function filters
   all theta \in theta_set which can be extended to theta' such that r and
   r' are isomorphic under theta'. 
   Returns set of all possibilities for theta'.                         *)
let match_rule_sets =
 match_termpair_sets (theta_set_for_rule_pair possible_theta)
;;

(********** match all ***************************************************)

(* check whether |e|=|e'| && |r|=|r'| && |c|=|c'|                       *)
let cardinalities_match (e, r, c) (e', r', c') =
 ((List.length e) == (List.length e')) &&
 ((Trs.size r) == (Trs.size r')) &&
 ((Trs.size c) == (Trs.size c'))
;;
(* Claim: if two processes (under the same selection startegy) have
   different lengths (i.e., oriented a different number of equations
   with non-unique orientation, they are not isomorphic 
   WRONG !!! see CGE *)
let lengths_match p p' = CP.length p = CP.length p'

let print_iso (p, (e, r, c)) (p', (e', r', c')) =
 Format.printf "The following two processes %s and %s are isomorphic:\n"
  (CP.to_string p) (CP.to_string p');
 Format.printf "Equations1:%s\n Rules1:%s\n Constraints1:%s\n"
  (Equation.all_to_string e) (Trs.to_string r) (Trs.to_string c);
 Format.printf "Equations1:%s\n Rules1:%s\n Constraints1:%s\n"
  (Equation.all_to_string e') (Trs.to_string r') (Trs.to_string c');
;;


(* the main function: gets two process states and checks whether their
   - current equation sets e, e' 
   - current rewrite systems r, r'
   - current constraint systems c, c'
  are equal under some isomorphism theta (all type Trs.t). 
  Boolean return value.                                                  *)
let processes_match (p, (e, r, c)) (p', (e', r', c')) =
 if 
  (* lengths match is not true !!! *)
  not ((cardinalities_match (e, r, c) (e', r', c')) (*&& lengths_match p p'*))
 then
  false
 else (
  if not (lengths_match p p') then print_iso (p, (e, r, c)) (p', (e', r', c'));
  let theta_set_0 = [SymSub.empty] in
  let theta_set = match_rule_sets r r' theta_set_0 in
  let theta_set' = match_rule_sets c c' theta_set in
  let theta_set'' = match_equation_sets e e' theta_set' in
  not (List.is_empty theta_set''))
;;


(*** TESTS ***************************************************************)

let test () =
 Format.printf "testing module TrsRenaming ... \n";
 let (x,y,z), (c,f,g,h,k), sigma = Tx.test_signature in
 let const = Term.Fun (c, []) in
 let f_x = Term.Fun (f, [x]) in
 let f_f_x = Term.Fun (f, [f_x]) in
 let h_h_c = Term.Fun (h, [Term.Fun (h,[const])]) in
 let h_h_x = Term.Fun (h, [Term.Fun (h, [x])]) in
 let h_x = Term.Fun (h, [x]) in
 let h_h_y = Term.Fun (h, [Term.Fun (h, [y])]) in
 let f_h_y = Term.Fun (f, [Term.Fun (h, [y])]) in
 let h_f_y = Term.Fun (h, [Term.Fun (f, [y])]) in
 let g_x_x = Term.Fun (g, [x; x]) in
 let g_x_y = Term.Fun (g, [x; y]) in
 let k_z_z = Term.Fun (k, [z; z]) in
 let g_f_x_h_h_x = Term.Fun (g, [f_x; h_h_x]) in
 let k_h_x_h_h_x = Term.Fun (k, [h_x; h_h_x]) in
 let r1 = Rule.of_terms f_f_x f_x in
 let r2 = Rule.of_terms h_h_x h_x in
 let r2' = Rule.of_terms h_h_y h_x in
 let r3 = Rule.of_terms f_h_y h_f_y in
 let r4 = Rule.of_terms g_x_x const in
 let trs = Trs.of_list [r1; r2; r4] in 
(*** Some asssertions ****************************************************)
 assert (terms_are_variants h_h_x h_h_y);
 assert (terms_are_variants h_h_x f_f_x);
 assert (terms_are_variants h_f_y f_h_y);
 assert (not (terms_are_variants h_h_c f_f_x));
 assert (not (terms_are_variants g_x_x f_x));
 assert (terms_are_variants g_x_x k_z_z);
 assert (not (terms_are_variants g_x_y k_z_z));
 assert (not (terms_are_variants g_f_x_h_h_x k_h_x_h_h_x));
 assert (not (rules_are_variants r1 r2'));
 assert (rules_are_variants r3 (Rule.invert r3));
 let b = 
 try 
  let _ = orientation_symmetric r3 trs in true 
 with Not_renamable -> false
 in
 assert b;
 let none = Trs.empty in
 let eqns1 = [Equation.of_terms g_x_x x] in
 let eqns2 = [Equation.of_terms k_z_z z] in
(* test if processes_match checks equations_match, symmetric *)
 let processes_match' p p' = processes_match ([], p) ([], p') in
 assert (processes_match' (eqns1, none, none) (eqns2, none, none));
 let eqns1' = [Equation.of_terms g_x_y x] in
(* nonlinearity substitution *)
 assert (not (processes_match' (eqns1', none, none) (eqns2, none, none)));
 let rs1 = Trs.add (Rule.of_terms g_x_x x) Trs.empty in
 let rs2 = Trs.add (Rule.of_terms k_z_z z) Trs.empty in
(* check if theta from eqns is verified in rules *)
 assert (processes_match' (eqns1, rs1, none) (eqns2, rs2, none));
 let cs1 = Trs.add (Rule.of_terms f_x x) rs1 in
 let cs2 = Trs.add (Rule.of_terms h_x x) rs2 in
(* check if theta is extended in rules *)
 assert (processes_match' (eqns1, rs1, cs1) (eqns2, rs2, cs2));
 let g_f_x_f_f_x = Term.Fun (g, [f_x; Term.Fun(f, [f_x])]) in
 let cs1' = Trs.add (Rule.of_terms k_h_x_h_h_x f_x) cs1 in
 let cs2' = Trs.add (Rule.of_terms g_f_x_f_f_x h_x) cs2 in
(* check if theta is extended in rules *)
 assert (processes_match' (eqns1, rs1, cs1') (eqns2, rs2, cs2'));
 assert (processes_match' (eqns1, cs1', rs1) (eqns2, cs2', rs2))
;;

(*test ();; *)
