(* Copyright 2010 Sarah Winkler
 * GNU Lesser General Public License
 *
 * This file is part of MKBtt.
 * 
 * MKBtt is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * MKBtt is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with MKBtt. If not, see <http://www.gnu.org/licenses/>.
 *)

(** Main file
@author Sarah Winkler
@since  2007/05/10 *)

open Util;;

(*** SUBMODULES **********************************************************)
module C = Completion;;
module W = World;;
module St = Statistics;;
module Monad = W.Monad;;
module NS = NodeState;;
module RMonad = U.Monad;;
module TrsParser = U.MyTrsParser;;
module TrsLexer = U.MyTrsLexer;;
module TrsSyntax = U.MyTrsSyntax;;
module Problem = U.MyProblem;;

(*** OPENS ***************************************************************)
open Monad;;
open Completion;;

(*** EXCEPTIONS **********************************************************)
exception Help;;
 
(*** GLOBALS *************************************************************)

let log0 = W.log 0

let log1 = W.log 1 

let log2 = W.log 2

let check_options =
 W.get_options >>= fun options ->
 let strategy_string = Completion.termination_strategy options in
 log2 ("Main check: strategy "^strategy_string)
;;

(* termination strategies *)
let s_linpoly = "(matrix -dim 1 -ib 1 -ob 4)+";;
let s_poly1 = "(poly -ib 1 -ob 4)+";;
let s_poly2 = "(poly -ib 2 -ob 4)+";;
let s_kbo = "(kbo -smt || ref;kbo)+";;
let s_lpo = "(lpo || ref;lpo)+";;
let s_total1 = "(kbo | lpo | matrix -dim 1 -ib 3 -ob 4)+";;
let s_total2 = "(kbo -smt || lpo || matrix -dim 1 -ib 3 -ob 3 || ref;kbo || ref;lpo || mpo)+";;

let s_ttt2micro = "dp;(edg | sccs | sc | matrix -ib 1)*";;

let s_ttt2fast = "dp;(edg | sccs | sc | lpo | matrix -ib 1)*";;

(* default settings *)
let flags = ref C.default_options


(***** functions to treat flags *****)

let set_cpc s =
 if (String.compare s "none") == 0 then 
  flags := {!flags with cpc = C.NoCPC} 
 else if (String.compare s "prime") == 0 then 
  flags := {!flags with cpc = C.Prime}
 else if (String.compare s "blocked") == 0 then 
  flags := {!flags with cpc =  C.Blocked}
 (*else if (String.compare s "connected") == 0 then 
  flags := {!flags with cpc =  C.Connected}
 else if (String.compare s "all") == 0 then 
  flags := {!flags with cpc =  C.All}*)
 else
  flags := {!flags with cpc = C.PrimeBlocked}
;;

let set_selection s =
 (*if (String.compare s "old") == 0 then
  (flags.selection_strategy := SelectionStrategy.old !(flags.size_age_ratio);
   flags.selection_macro := Some(SelectionStrategy.Old))
 else*) if (String.compare s "sum") == 0 then
  flags := {!flags with 
    selection_strategy = SelectionStrategy.japan_sum;
    strategy_macro = Some(SelectionStrategy.JapanSum)}
 else if (String.compare s "max") == 0 then
  flags := {!flags with  
    selection_strategy = SelectionStrategy.japan_max;
    node_size = C.Max;
    strategy_macro = Some(SelectionStrategy.JapanMax)}
 else 
  (flags := {!flags with strategy_macro = None};
   if (String.compare s "slothrop") == 0 then
    flags := {!flags with selection_strategy = SelectionStrategy.slothrop}
  else
   let strategy = SelectionStrategyAux.of_string s in
   flags := {!flags with selection_strategy = strategy}
  )
;;

let set_iso s =
 if (String.compare s "none") == 0 then
  flags := {!flags with check_isomorphism_with = C.NoChecks}
 else if (String.compare s "rename") == 0 then 
  flags := {!flags with check_isomorphism_with = C.Renamings} 
 else if (String.compare s "rename+") == 0 then
  flags := {!flags with 
   check_isomorphism_with = C.Renamings;
   check_isomorphisms_repeatedly = true}
 else if (String.compare s "perm") == 0 then 
  flags := {!flags with check_isomorphism_with = C.Permutations}
 else if (String.compare s "perm+") == 0 then
  flags := {!flags with
   check_isomorphism_with = C.Permutations;
   check_isomorphisms_repeatedly = true}
;;

let set_rindex i = flags := {!flags with rewrite_index = i}

let set_uindex i = flags := {!flags with unification_index = i}

let set_index set s =
(* if (String.compare s "dt") == 0 then set C.DiscriminationTrees
 else if (String.compare s "pi") == 0 then set C.PathIndexing
 else*) if (String.compare s "ct") == 0 then set C.CodeTrees
 else set C.Naive
;;

let set_strategy s =
 if (String.compare s "slinpoly") == 0 then
  flags := {!flags with termination_strategy = s_linpoly}
 else if (String.compare s "spoly1") == 0 then
  flags := {!flags with termination_strategy = s_poly1}
 else if (String.compare s "spoly2") == 0 then
  flags := {!flags with termination_strategy = s_poly2}
 else if (String.compare s "slpo") == 0 then
  flags := {!flags with termination_strategy = s_lpo}
 else if (String.compare s "skbo") == 0 then
  flags := {!flags with termination_strategy = s_kbo}
 else if (String.compare s "ttt2micro") == 0 then
  flags := {!flags with termination_strategy = s_ttt2micro}
 else if (String.compare s "ttt2fast") == 0 then
  flags := {!flags with termination_strategy = s_ttt2fast}
 else if (String.compare s "stotal1") == 0 then
  flags := {!flags with termination_strategy = s_total1}
 else if (String.compare s "stotal2") == 0 then
  flags := {!flags with termination_strategy = s_total2}
 else flags := {!flags with termination_strategy = s};
  flags := {!flags with check_externally = false}
;;


let set_node_size s =
 if (String.compare s "max") = 0 then 
  flags:= {!flags with node_size = C.Max}
 else
  flags:= {!flags with node_size =  C.Sum}
;;

let set_cmd c =
 flags := {!flags with tool = "./" ^ c; check_externally = true}
;;

let set_completion_check c =
 flags := {!flags with mode = CompletenessCheck(c)}
;;

let set_output_trs () = flags := {!flags with output_trs = true}

let set_output_proof () = flags := {!flags with output_proof = true}

let set_output_stats () = flags := {!flags with output_statistics = true}

let set_ordered () = flags := {!flags with ordered_completion = true}

let set_slothrop () = flags := {!flags with slothrop = true}

let set_timeout f = flags := {!flags with timeout = f}

let set_ttimeout f = flags := {!flags with ttimeout = f}

let set_verbosity v = flags := {!flags with verbosity = v}

(* Available command line switches. *)
let sws = [
 ("-ch", Arg.String set_completion_check,
  "supply rewrite system to check whether it is complete and joins the input equalities");
 ("-cp", Arg.String set_cpc,
  "critical pair criterion (prime, connected, blocked, primeblocked, connected or all)");
 ("-ct", Arg.Unit set_output_trs,
  "output completed TRS");
 ("-help", Arg.Unit (fun () -> raise Help),
  "display this list of options");
 ("-h", Arg.Unit (fun () -> raise Help),
  "display this list of options");
 ("-is", Arg.String set_iso,
  "isomorphism checks (rename or perm, possibly extended with +)");
 ("-ix", Arg.String (set_index set_rindex),
  "indexing technique for rewriting (nv, ct)");
(*  "indexing technique for rewriting (nv, pi, dt, ct)");*)
(* ("-ns", Arg.String set_node_size,
  "max or sum (default)");*)
 ("-p", Arg.Unit set_output_proof,
  "output proof");
 ("-ss", Arg.String set_selection,
  "selection strategy (sum, max, old and slothrop are predefined)");
(* ("-o", Arg.Unit set_ordered,
  "ordered completion");*)
 ("-s", Arg.String set_strategy,
  "termination strategy for TTT2, e.g. kbo, spoly1, stotal2");
(* ("-sa", Arg.Set_float flags.size_age_ratio,
  "ratio of size (instead of age) choices");*)
(* ("--slothrop", Arg.Unit set_slothrop,
  "Use Slothrop.");*)
 ("-st", Arg.Unit set_output_stats,
  "output statistics");
 ("-t", Arg.Float set_timeout,
  "global time limit");
 ("-T", Arg.Float set_ttimeout,
  "termination call time limit");
 ("-tp", Arg.String set_cmd,
  "executable script for calling external termination prover");
 ("-ui", Arg.String (set_index set_uindex),
  "indexing technique for overlaps (nv, pi, dt)");
(* ("-v", Arg.Int set_verbosity,
  "verbosity in {0, 1, 2}");*)
(* ("--waldgoal", Arg.Set flags.use_wald_goal,
  "use Waldmeister-like goal representation");*)
]

(* Align the help message nicely. *)
let sws = Arg.alignx 80 sws;;

(*** FUNCTIONS ***********************************************************)
(*let usage cmd = Format.sprintf "Usage: %s [options] <file>@\n" cmd;;*)
let usage = "Usage: ./mkbtt [options] <file> [timeout]\n\nOptions:";;


let configurate argv =
 if Array.length argv = 1 then (Arg.usagex sws usage; exit 0);
 let file_is_set = ref true in
 let push_file n = 
  if !file_is_set then
   (file_is_set := false; 
   flags := {!flags with filename = n})
  else
   (Arg.usagex sws usage; exit 0)
 in
 Arg.parse sws push_file usage;
;;

let if_do b m = if b then m else return ()

let output (status, result) t =
 let problem_name = Filename.basename !flags.filename in
 St.print_basic status problem_name t; 
 begin
 match Completion.mode (!flags),status,result with
  | C.Completion,C.Sat,Some(C.SlothropResult r) ->
   if_do !flags.output_trs  (St.print_system r []) >>
   let e0 = !flags.axioms in
   if_do !flags.output_statistics (St.print_success e0 [] t)
  | C.Completion,C.Sat,Some(C.MKBttResult p) ->
   if_do !flags.output_proof (Control.print_proof status p) >>
   NS.project_r_closed p >>= fun r ->
   NS.project_e_closed p >>= fun e ->
   if_do !flags.output_trs  (St.print_system r e) >>
   let e0 = !flags.axioms in
   if_do !flags.output_statistics (St.print_success e0 p t)
  | C.Completion, C.Gup _, _
  | C.Completion, C.Unk, _
  | C.Completion, C.Tmo, _ ->
   if_do !flags.output_statistics (St.print_timeout t)
  | C.Proof, C.Sat, Some (C.MKBttResult p)
  | C.Proof, C.Uns, Some (C.MKBttResult p)
  | C.Proof, C.Noc, Some (C.MKBttResult p)
  | C.Proof, C.Suc, Some (C.MKBttResult p) ->
   if_do !flags.output_proof (Control.print_proof status p) >>
   let e0 = !flags.axioms in
   if_do !flags.output_statistics (St.print_proof status e0 p t)
  | C.Proof, C.Unk, _
  | C.Proof, C.Gup _, _
  | C.Proof, C.Tmo, _ -> 
   if_do !flags.output_statistics (St.print_timeout t)
  | C.CompletenessCheck _, C.Suc,  _ -> 
   log0 "Completeness and joinability of axioms verified."
  | C.CompletenessCheck _, C.Noc, _ ->
   log0 "Completeness and joinability of axioms not verified."
  | _ -> raise (C.GiveUp "Strange result state")
 end >> St.print_logs
;;

let goal_is_empty = function
 | C.Universal s
 | C.Existential s  -> List.is_empty s
;;

let read_wst_file filename so = 
 let chin = Pervasives.open_in filename in
 let lbuf = Lexing.from_channel chin in
 let ts = TrsParser.trs TrsLexer.token lbuf in
 let empty = U.Signature.empty 100 in
 let s = match so with Some s -> s | None -> empty in
 let p, s = TrsSyntax.to_problem_with ts s in
 liftm (RMonad.set s) >>
 let trs = Problem.get_trs p in
 return trs
;;

let read_file f =
 if Filename.check_suffix f "tptp" then
  liftm (TPTPInput.of_file f) >>= fun (eqs, g, ck) ->
   flags := {!flags with axioms = eqs};
   W.log 2 ("Read TPTP file"^f) >>
   if !flags.ordered_completion && (not (goal_is_empty g)) then (
     flags := {!flags with  
      goal = g; conjecture = Some (ck); mode = C.Proof};
     return (eqs, Some g))
   else
    return (eqs, None)
  else
   read_wst_file f None >>= fun trs ->
   (*W.log 2  ("Read TRS file "^f) >>*)
   let eqs = Trsx.to_equations trs in
   flags := {!flags with axioms = eqs};
   return (eqs, None)
;;

let run_completion (eqs, goal) =
 W.M.Equation.list_to_stringm eqs >>= fun s ->
 if not (!flags.slothrop) then
   Control.complete eqs goal
 else
  SlMain.go eqs >>= fun trs ->
  return (C.Suc, Some (C.SlothropResult trs))
;;

let check_completion eqs trs_filename =
 liftm RMonad.get >>= fun s ->
 read_wst_file trs_filename (Some s) >>= fun trs ->
 W.M.Trsx.to_stringm trs >>= fun s ->
 W.M.Equation.list_to_stringm eqs >>= fun s ->
 Check.termination trs >>= fun is_terminating ->
 Check.confluence trs >>= fun is_confluent ->
 Check.join_axioms eqs trs >>= fun joins_axioms ->
 if not is_terminating then
  (log1 "System was not shown terminating." >> return (C.Noc, None))
 else if not is_confluent then
  (log1 "System is not confluent." >> return (C.Noc, None))
 else if not joins_axioms then
  (log1 "System does not join input equalities." >> return (C.Noc, None))
 else return (C.Suc, None)
;;

let work () = 
 (read_file !flags.filename >>= fun input ->
  match !flags.mode with
   | C.CompletenessCheck cs -> check_completion (fst input) cs
   | _ -> run_completion input ) >>= fun r ->
 get >>= fun s' -> liftm RMonad.get >>= fun s -> return (r, s, s')
;;

let work_nonmonadic (s,s') =
 let r = 
 (* try*) RMonad.run s (run s' (work ())) (*with
  Failure e -> (Format.eprintf "%s@\n%!" e; failwith e)*)
 in
 if Either.is_right r then Either.right r 
 else failwith "work: no result found"
;;

let go () =
 liftm RMonad.get >>= fun s -> get >>= fun s' ->
 let t = !flags.timeout in
 log2 ("Timeout of "^(string_of_float t)) >>
 match
  Process.run_timed (Process.Local t) work_nonmonadic (s,s')
 with
  | None,_ -> output (C.Tmo, None) 0.0
  | Some (x,s,s'),r -> liftm (RMonad.set s) >> set s' >> output x (t -. r)
;;


let execute () =
 try
  let start = Unix.gettimeofday () in
  configurate Sys.argv;
  let s = U.Signature.empty 200 in
  let s' = W.initial_context_with !flags in
  let r = RMonad.run s (W.Monad.run s' (go ())) in
  let t = (Unix.gettimeofday ()) -. start in
  Format.printf "Total time: %f\n" t;
  if Either.is_right r then Either.right r
  else failwith "No result found."
 with
 (* | Failure s -> Format.eprintf "%s@\n%!" s*)
  | Help -> Arg.usagex sws usage;
;;

execute ()
