(* Copyright 2010 Sarah Winkler
 * GNU Lesser General Public License
 *
 * This file is part of MKBtt.
 * 
 * MKBtt is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * MKBtt is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with MKBtt. If not, see <http://www.gnu.org/licenses/>.
 *)

(** Data type and functions related to goal in theorem proving
    This implementation uses just nodes to represent the conjecture.
@author Sarah Winkler
@since  2009/07/03 *)

(*** OPENS (1) ***********************************************************)
open Util;;

(*** SUBMODULES **********************************************************)
module C = Completion;;
module Pos = Rewriting.Position;;
module Rule = U.Rule;;
module W = World;;
module Monad = W.Monad;;
module GS = GoalState;;
module NS = NodeState;;
module NSA = NodeSetAux;;
module NI = NodeTermIndexx;;
module N = IndexedNode;;
(*** OPENS (2) ***********************************************************)
open Monad;;

(*** FUNCTIONS ***********************************************************)

(* ------------------------------------------------------------------ *)
(*  CREATE goal                                                       *)
(* ------------------------------------------------------------------ *)
let dummy = []

(* gets skolemized equation, returns goal *)
let set cs = 
 let eqns_of c =
  let s, t = Equation.terms c in
  GS.equals_x_x_is_true >>= fun e ->
  GS.equals_s_t_is_false s t >>= fun e' ->
  NSA.of_axioms [e] >>= fun ns_true ->
  NSA.of_axioms [e'] >>= fun ns_false ->
  NS.add_open ns_true >>
  GS.set_ngoal ns_false
 in
 match cs with
 | [c] -> eqns_of c
 | _ ->  raise (C.GiveUp "Goal module supports only single goal.") 
;;

let to_string = 
 GS.get_ngoal >>= fun g ->
 NSA.to_string g
;;

(* ------------------------------------------------------------------ *)
(*  REWRITE goal                                                      *)
(* ------------------------------------------------------------------ *)

let union = List.union

let union3 x y = union (union x y)

let rewrite_node (retrieve, el) (n,b) single_index =
 N.brule b n >>= fun rule ->
 retrieve single_index (Rule.rhs rule) >>= fun matches ->
 let add cn m =
  MKBtt.rewrite_node_with true el (n,b) m >>= fun c -> return (union c cn)
 in
 foldl add [] matches
;;

let rewrite_goal fs gs idx =
 let add_reducts (irn, cn) n =
  rewrite_node fs (n,true) idx >>= fun cn_lr ->
  rewrite_node fs (n,false) idx >>= fun cn_rl ->
  let cn' = union cn_lr cn_rl in
  if List.is_empty cn' then return (union [n] irn,cn)
  else return (irn,union cn' cn)
 in
 foldl add_reducts ([], []) gs
;;

(* rewrite1 uses generic rewrite function, with appropriate index
   retrieval and equation label filtering parameters *)
let rewrite1 =
 let pair_root x = return (x,Pos.root) in
 let retrieve idx t = NI.variant_candidates_in idx t >>= map pair_root in
 let elabels l2 l3 = l3 in
 rewrite_goal (retrieve, elabels)
;;

(* rewrite2 uses generic rewrite function, with appropriate index
   retrieval and equation label filtering parameters *)
let rewrite2 =
 let retrieve = NI.encompassment_candidates_in in
 let elabels l2 l3 = union l2 l3 in
 rewrite_goal (retrieve, elabels)
;;

let rewrite_once_with single_index =
 GS.get_ngoal >>= fun g ->
 rewrite1 g single_index >>= fun (_, created1) ->
 rewrite2 g single_index >>= fun (_, created2) ->
 MKBtt.gc (union created1 created2)
;;

let rewrite _ single_index =
 rewrite_once_with single_index >>= fun g' ->
 if List.is_empty g' then 
  return ()
 else
  GS.get_ngoal >>= fun g ->
  MKBtt.rewrite_open g' >>= fun g'' ->
  MKBtt.gc (union3 g g' g'') >>=
  GS.set_ngoal
;;

(* ------------------------------------------------------------------ *)
(*  CHECK whether goal is proven                                      *)
(* ------------------------------------------------------------------ *)

let proven_for =
  GS.true_is_false >>= fun eq ->
  N.by_eqn eq >>= function
   | Some n -> N.e_all n
   | None -> return []
;;

