(* Copyright 2010 Sarah Winkler
 * GNU Lesser General Public License
 *
 * This file is part of MKBtt.
 * 
 * MKBtt is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * MKBtt is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with MKBtt. If not, see <http://www.gnu.org/licenses/>.
 *)

(** check termiantion with external tool 
@author Sarah Winkler
@since  2008/02/19
*)

(** Calls external termination checker *)

(*** OPENS ********************************************************************)
open Util;;

(*** SUBMODULES **********************************************************)
module Var = Rewriting.Variable;;
module Fun = Rewriting.Function;;
module Term = U.Term;;
module Rule = U.Rule;;
module Trs = U.Trs;;
module Signature = U.Signature;;
module W = World;;
module Monad = W.Monad;;

(*** OPENS (2) ***********************************************************)
open World;;

(*** TYPES ***************************************************************)
type outcome = Result of bool | TimedOut | Unknown

(*** FUNCTIONS ***********************************************************)
let (>>=) = Monad.(>>=)
let (>>) = Monad.(>>)
let return = Monad.return

(*
let rec term_to_string = function
 | Term.Var x -> Var.to_string x
 | Term.Fun(f, ts) -> Fun.to_string f ^ "(" ^ (termlist_to_string ts) ^ ")"
and termlist_to_string = function
 | [] -> ""
 | [t] -> term_to_string t
 | t1 :: ts -> (term_to_string t1) ^ ", " ^ (termlist_to_string ts)
;;*)



let weak_rule_to_string rule =
 let l, r = Rule.lhs rule, Rule.rhs rule in
 let ls, rs = Term.to_string l, Term.to_string r in
 ls ^ " ->= " ^ rs
;;

let rule_to_string rule =
 let l, r = Rule.lhs rule, Rule.rhs rule in
 let ls, rs = Term.to_string l, Term.to_string r in
 ls ^ " -> " ^ rs
;;


let rules_to_string rules f = 
 Trs.fold (fun r s -> (f r) ^ "\n" ^ s) "" rules
;;

let strict_rules_to_string rules = 
 rules_to_string rules rule_to_string
;;

let weak_rules_to_string rules =
 rules_to_string rules weak_rule_to_string
;;

(* gives a WST string format for trs, where strict is true iff '->' is
   used, otherwise '->=' *)
let varstring trs =
 let tvars = Term.vars in
 let vars = 
  Trs.fold (fun r v -> (tvars (Rule.rhs r)) @ v @ (tvars (Rule.lhs r))) [] trs
 in
 let s = List.fold_left (fun s x -> (Var.to_string x) ^ " " ^ s ) "" vars in
 "(VAR " ^ s ^ ")"
;; 
 
let trs_to_string trs = 
 (varstring trs) ^ "\n(RULES \n" ^ (strict_rules_to_string trs) ^ ")"
;;

let dp_problem_to_string dp trs =
 let rs = Trs.union dp trs in
 let dps, trss = strict_rules_to_string dp, weak_rules_to_string trs in
 let rulestr =  "\n (RULES \n" ^ dps ^ "\n" ^ trss ^ ")" in
 (varstring rs) ^ rulestr
;;


let read_from indescr =
 let line = String.make 3 ' ' in
 let l = Unix.read indescr line 0 3 in
 let line = String.sub line 0 l in
 (String.compare line "YES") == 0
;;

let run_external cmd s timeout = 
 let rfd0, wfd0 = Unix.pipe () in (* comm ocaml -> process *)
 let rfd1, wfd1 = Unix.pipe () in (* comm process -> ocaml *)
 let finish pid =
    (try Unix.kill pid 1 with _ -> ());
    (try ignore (Unix.waitpid [] pid) with _ -> ());
    (try Unix.close rfd1 with _ -> ()) in
  (* write in wfd0 *)
  let l = String.length s in
  let l' = Unix.write wfd0 s 0 l in (* l' ... chars written *)
  assert(l' = l);
  flush (Unix.out_channel_of_descr wfd0);
  Unix.close wfd0;
  let pid = Unix.create_process cmd (Array.of_list [])
    rfd0 wfd1 (Unix.descr_of_out_channel stderr)
  in Unix.close wfd1; Unix.close rfd0;
  let run () =
   try
    let outpt = read_from rfd1 in finish pid; Result (outpt)
   with
    | End_of_file -> finish pid; Unknown
  in
  match Process.run_timed (Process.Local timeout) run () with
   | None, t -> finish pid; (Unknown, t)
   | Some r, t -> r, t
;;

(*
let write_to_tmp s =
 (* Open temp file *)
 let name = Filename.temp_file "mkbtt" ".trs" in
 let descr = Unix.openfile name [Unix.O_RDWR] 0o600 in
 (* Write output *)
 let out_channel = Unix.out_channel_of_descr descr in
 Printf.fprintf out_channel "%s\n" s;
 flush out_channel;
 (* Close the underlying file descriptor, return tmp file name *)
 Unix.close descr;
 name
;;

let remove_tmp_file name =
 Sys.remove name
;;

let read_and_store indescr =
 let line = String.make 3 ' ' in
 let l = Unix.read indescr line 0 3 in
 let line = String.sub line 0 l in
 if ((String.compare line "YES") == 0) then
  current_result := Result true
 else
  current_result := Result false
;;

let print_errs indescr =
 let line = String.make 20 ' ' in
 let _ = Unix.read indescr line 0 20 in
 current_result := Unknown
;;

(* safe, but with files *)
let run_external2 cmd s timeout =
 let filename = write_to_tmp s in
 let args = Array.of_list [cmd; filename] in
 let (time, status) =
  Process.execute_timed timeout cmd args read_and_store print_errs
 in
 remove_tmp_file filename;
 (!current_result, time)
;;

(* The following functions use pipes - apparently not safe!

(* execute an external program with a certain timeout *)
let rec wait p_in p_er read_er () =
 match Unix.select [p_in;p_er] [] [] ~-.1.0 with
  | (p,[],[]) ->
   if List.mem p_er p then read_er p_er;
   if not (List.mem p_in p) then wait p_in p_er read_er ()
  | _ -> raise (Invalid_argument "Process.wait")
;;

let my_execute_timed t cmd s =
 let (my_out, tool_in) = Unix.pipe () in
 let (tool_out, my_in) = Unix.pipe () in
(* fill pipe *)
 let l = String.length s in
 let l' = Unix.write tool_in s 0 l in (* l' ... chars written *)
 assert(l' = l);
 flush (Unix.out_channel_of_descr tool_in);
 Unix.close tool_in;
 let number = inc () in
(* pipe filled *)
 let pid = Unix.fork () in
 if pid = 0 then ( (* child *)
 Printf.fprintf stderr "child %i\n" number; flush stderr;
  let path = Filename.dirname cmd in
  let cmd = "./" ^ Filename.basename cmd in
  Unix.chdir path;
  ignore (Unixx.setpgrp ());
  Unix.close tool_out;
  Printf.fprintf stderr "before %i - " number; flush stderr;
  let pid = Unix.create_process cmd [|cmd|] my_out my_in Unix.stderr in
(* close stuff? *)
  Printf.fprintf stderr "finished with %i (pid %i)" number pid;
(*  Unix.close stderr;*)
  flush (Unix.out_channel_of_descr my_in);
  Unix.close my_out;
  Unix.close my_in;
(* *)
  match Unix.waitpid [] pid with
   | (_,Unix.WEXITED s) -> Printf.fprintf stderr "exit s\n"; exit s
   | _                  -> Printf.fprintf stderr "exit 1\n"; exit 1
 ) else try (* parent *)
 Printf.fprintf stderr "parent %i\n" number;
  Unix.close my_out;
  Unix.close my_in;
  Printf.fprintf stderr "Waiting for %i (pid %i)\n" number pid;
(*  Unix.close stderr;*)
  let (_, rest) = Process.run_timed t (wait tool_out Unix.stderr print_errs) in
  read_and_store tool_out;
  Unix.close tool_out;
(*  Unix.close p_er;*)
  match Unix.waitpid [] pid with
   | (_,Unix.WEXITED s) -> (t -. rest,s)
   | _                  -> (t -. rest,1)
 with
 | Process.Timeout ->
  Format.printf "%i timed out\n" number; flush stdout;
  Unix.kill pid Sys.sigkill;
  Unix.close tool_out;
(*  Unix.close p_er;*)
  ignore (Unix.waitpid [] pid);
  (t,0)
;;

(* safe with pipes *)
let run_external3 cmd s timeout =
 let (time, status) =
  my_execute_timed timeout cmd s
 in
 (!current_result, time)
;;
*)
*)

let check trs =
 W.get_options >>= fun options ->
 Statistics.inc_n_termination_calls >>
 let s = trs_to_string trs in
 (* important to add ./ , otherwise command not found - no error :( *)
 let cmd = "./" ^ Completion.command options in
 let timeout = Completion.ttimeout options in
(* let s = trs_to_string trs in*)
 let (result, time) = run_external cmd s timeout in
 let dtime = timeout -. time in
 Statistics.add_t_termination dtime >>
 match result with
  | Result(true) ->
   Statistics.inc_n_termination_yes >> return true
  | TimedOut ->
   Statistics.inc_n_termination_timeouts >> return false
  | _ -> return false
;;

    
