(* Copyright 2008 Martin Korp, Christian Sternagel, Harald Zankl
 * GNU Lesser General Public License
 *
 * This file is part of TTT2.
 * 
 * TTT2 is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * TTT2 is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with TTT2. If not, see <http://www.gnu.org/licenses/>.
 *)

(*** OPENS ********************************************************************)
open Util;;
open Rewritingx;;

(*** MODULES ******************************************************************)
module C = Complexity;;
module F = Format;;
module Fun = Function;;
module M = Monad;;
module Number = Logic.Number;;
module P = Problem;;
module Var = Variable;;

(*** TYPES ********************************************************************)
type flags = {
  a : bool ref;
  curry : bool ref;
  help : bool ref;
  mirror : bool ref;
  u1 : bool ref;
  u2 : bool ref;
 };;
type head = Var of Var.t | Fun of Fun.t;;
type t = P.t * P.t;;

(*** GLOBALS ******************************************************************)
let code = "uncurry";;
let name = "Uncurry Processor";;
let keywords = ["uncurrying";"transformation"];;
let comment = "Implements uncurrying for applicative systems.";;
let flags = {
  a = ref false;
  curry = ref false;
  help = ref false;
  mirror = ref false;
  u1 = ref false;
  u2 = ref false;
};;

let spec =
 let spec = [
  ("--help",Arg.Set flags.help,"Prints information about flags.");
  ("-help",Arg.Set flags.help,"Prints information about flags.");
  ("-h",Arg.Set flags.help,"Prints information about flags.");
  ("-a",Arg.Set flags.a,"Transformation A.");
  ("-curry",Arg.Set flags.curry,"Curry the system.");
  ("-mirror",Arg.Set flags.mirror,"Mirrors the TRS before it is uncurried.");
  ("-u1",Arg.Set flags.u1,"processor U1.");
  ("-u2",Arg.Set flags.u2,"processor U2.");
  ] in
 Arg.alignx 80 spec
;;

let help = (comment,keywords,List.map Triple.drop_snd spec);;

(*** FUNCTIONS ****************************************************************)
let (>>=) = M.(>>=);;
let (>>) = M.(>>);;
let init _ = 
 flags.a := false;
 flags.help := false;
 flags.mirror := false;
 flags.u1 := false;
 flags.u2 := false;
;;

(* Processor *)
let max_ari trs =
 List.foldr (M.lift2 max <.> M.find_ari) (M.return ~-1) (Trs.funs trs)
;;

(*let n_vars trs = max_ari trs >>= fun ma -> M.return (ma+1);;*)

let fresh_var = M.fresh_var >>= (M.return <.> Term.make_var);;

let uncurry app f a = 
 if a = 0 then M.return f else M.set_curry ~a:a f a >>= fun f ->
 M.is_dp app >>= fun b -> if b then M.set_dp f else M.return f
;;

let drop_curry f = M.is_curry f >>= fun b -> if b then M.drop_curry f else M.return f;;
let curry app f = 
 M.is_dp f >>= fun b -> if b then M.set_dp app else M.return app
;;

let is_extended_applicative atrs = 
 let fs = Trs.funs atrs in
 M.filter (M.is_ari 0) fs >>= fun cs ->
 M.filter (M.is_ari 2) fs >>= fun bs ->
 M.return (List.equal fs (cs@bs))
;;

let is_dp_applicative p =
 let (s,w) = P.get_sw p in
 is_extended_applicative s >>= fun b ->
 is_extended_applicative w >>= fun c ->
 M.return (b && c)
;;

let app trs = 
 Trs.is_applicative trs >>= fun b ->
 if b then () else failwith "TRS not applicative";
 M.lift List.hd (M.filter (M.is_ari 2) (Trs.funs trs))
;;

let app_dp trs = 
 is_extended_applicative trs >>= fun b ->
 if b then () else failwith "TRS not extended applicative";
 M.filter (M.is_ari 2) (Trs.funs trs)
;;

let rec head = function 
 | Term.Var x -> Var x
 | Term.Fun (f,[]) -> Fun f
 | Term.Fun (f,[s;_]) -> head s
 | _ -> failwith "term not applicative"
;;

(* applicative arity *)
let rec aa_t = function
 | Term.Var _ -> 0
 | Term.Fun (_,[]) -> 0
 | Term.Fun (_,[t;_]) -> 1 + aa_t t 
 | _ -> failwith "term not applicative"
;;

let aa_term f t = 
 let equal h f = match (h,f) with Fun f,g -> f = g | _ -> false in
 let ts = List.filter (fun t -> equal (head t) f) (Term.subterms t) in
 List.foldr max 0 (List.map aa_t ts)
;;

let aa_rule f rule = 
 let (l,r) = Pair.map (aa_term f) (Rule.to_terms rule) in
 List.filter (fun aa -> aa >= 0) [l;r]
;;

(* applicative arity for [f] *)
let aa_f f trs = 
 let aa = List.flat_map (aa_rule f) (Trs.to_list trs) in
 List.foldr max 0 aa
;;

let pick_arguments t =
 let rec pick_arguments acc = function
  | Term.Var _ -> failwith "head variable"
  | Term.Fun (_,[]) -> acc
  | Term.Fun (_,[s;t]) -> pick_arguments (t :: acc) s 
  | _ -> failwith "term not applicative"
 in
 pick_arguments [] t
;;

let rec proper_t trs t =
 match head t with
  | Var _ -> true
  | Fun f ->
   let aa = aa_f f trs in
   let args = pick_arguments t in
   List.length args = aa && List.for_all (proper_t trs) args
;;

let proper trs = M.return (List.for_all (proper_t trs) (Trs.terms trs));;

(* applicative arity for [t] *)
let rec aa_t t trs = match t with
 | Term.Var _ -> failwith "non left-head-variable-free term"
 | Term.Fun (f,[]) -> aa_f f trs 
 | Term.Fun (f,[t;_]) -> aa_t t trs - 1
 | _ -> failwith "term not in applicative form"
;;

(* eta saturation process *)
let eta_sat_rule trs r = 
 let (l,r) = Rule.to_terms r in
 let aa_t = aa_t l trs in
 let rec add f l r aa_t acc =
  if aa_t <= 0 then acc
  else
   fresh_var >>= fun x -> acc >>= fun acc ->
   let l' = Term.Fun (f,[l;x]) and r' = Term.Fun (f,[r;x]) in
   add f l' r' (aa_t - 1) (M.return (Rule.of_terms l' r' :: acc))
 in
 app trs >>= fun f -> add f l r aa_t (M.return [])
;;

let eta_sat trs =
 M.lift Trs.of_list (M.flat_map (eta_sat_rule trs) (Trs.to_list trs))
;;

let rec head_variable_free t = match head t with
 | Var _ -> false
 | Fun _ -> true
;;

let head_variable_free ts = 
 let subterms = Term.subterms ~p:(not <.> Term.is_var) in
 let st = List.flat_map subterms ts in
 List.for_all head_variable_free st
;;

let left_head_variable_free trs = head_variable_free (Trs.lhs trs);;

let head_variable_free trs = head_variable_free (Trs.lhs trs @ Trs.rhs trs);;

(* uncurry symbols in [fs] of term [t] *)
let rec ff_term fs trs t = match t with
 | Term.Var _ -> M.return t
 | Term.Fun (_,[]) -> M.return t
 | Term.Fun (app,[u;v]) when List.mem app fs ->
  (match head t with 
   | Var _ ->
    M.lift2 Term.make_fun (M.return app)
     (M.sequence [ff_term fs trs u;ff_term fs trs v])
   | Fun f -> 
    let aa = aa_f f trs and args = pick_arguments t in
    let a = List.length args in 
    if a > aa then
     M.lift2 Term.make_fun (M.return app)
      (M.sequence [ff_term fs trs u;ff_term fs trs v])
    else 
     M.lift2 Term.make_fun (uncurry app f a) (M.map (ff_term fs trs) args))
 | Term.Fun (f,[u;v]) ->
  M.lift2 Term.make_fun (M.return f) (M.map (ff_term fs trs) [u;v])
 | t -> failwith "term not applicative"
;;

let ff_rule fs trs rule = 
 let (l,r) = Rule.to_terms rule in
 ff_term fs trs l >>= fun l -> ff_term fs trs r >>=
 (M.return <.> Rule.of_terms l)
;;

let ff fs trs trs2 =
 M.lift Trs.of_list (M.sequence (Trs.map (ff_rule fs trs) trs2))
;;


let rec aux_rules_f var_list app f aa =
 if aa < 0 then
  M.return Trs.empty
 else (
  let vars = List.take (aa + 1) var_list in
  assert (List.length vars > 0);
  let y, var = List.hd vars, List.tl vars in
  let i = List.length var in
  uncurry app f i >>= fun fi ->
  uncurry app f (i+1) >>= fun fi' ->
  let l = Term.Fun (app, [Term.Fun (fi, var); y]) in
  let r = Term.Fun (fi', var @ [y]) in
  (aux_rules_f var_list app f (aa - 1)) >>= fun aux ->
  M.return (Trs.add (Rule.of_terms l r) aux))
;;

let aux_rules var_list atrs =
 app atrs >>= fun app ->
 M.filter (M.is_ari 0) (Trs.funs atrs) >>=
 List.foldr
  (fun f -> M.lift2 Trs.union (aux_rules_f var_list app f (aa_f f atrs - 1)))
  (M.return Trs.empty)
;;

let transform aa atrs = 
 if Trs.is_empty atrs then M.return None else
 Trs.is_applicative atrs >>= fun c ->
 if c then 
  if left_head_variable_free atrs then
   let fs = Trs.funs atrs in
   ff fs atrs atrs >>= fun trs -> max_ari trs >>= 
   (M.lift List.rev <.> M.sequence <.> List.gen (const fresh_var)) >>=
   flip aux_rules atrs >>= fun aux -> eta_sat atrs >>= ff fs atrs >>=
   (M.return <.> Option.some <.> Trs.union (Trs.union trs aux))
  else M.return None
 else M.return None
;;


(* new begin *)

let rec uncurried_term aa apps cs t = match t with
 | Term.Var _ -> M.return t
 | Term.Fun (_,[]) -> M.return t
 | Term.Fun (app,[u;v]) when List.mem app apps ->
  (match head t with 
   | Var _ ->
    M.lift2 Term.make_fun (M.return app)
     (M.sequence [uncurried_term aa apps cs u;uncurried_term aa apps cs v])
   | Fun f -> 
    let aaf = aa f and args = pick_arguments t in
    let a = List.length args in 
    if a > aaf then
     M.lift2 Term.make_fun (M.return app)
      (M.sequence [uncurried_term aa apps cs u;uncurried_term aa apps cs v])
    else 
     M.lift2 Term.make_fun (uncurry app f a) (M.map (uncurried_term aa apps cs) args))
 | Term.Fun (f,[u;v]) ->
  M.lift2 Term.make_fun (M.return f) (M.map (uncurried_term aa apps cs) [u;v])
 (* | t -> failwith "term not applicative" *)
 | t -> M.return t
;;

let uncurried_rule aa apps cs rule =
 let (l,r) = Rule.to_terms rule in
 uncurried_term aa apps cs l >>= fun l -> uncurried_term aa apps cs r >>= fun r ->
 M.return (Rule.of_terms l r)
;;

let uncurried_system aa apps cs trs =
 M.lift Trs.of_list (M.sequence (Trs.map (uncurried_rule aa apps cs) trs))
;;

let rec curried_term app = function
 | Term.Var x as v -> M.return v
 | Term.Fun (f,[]) -> drop_curry f >>= fun f -> M.return (Term.Fun(f,[]))
 | Term.Fun (f,ts) -> 
  drop_curry f >>= fun f ->
  List.foldl (fun acc t -> 
   acc >>= fun acc' ->
   curried_term app t >>= fun t' ->
   M.return (Term.Fun (app,[acc';t']))
  ) (M.return (Term.Fun (f,[]))) ts
;;

let sharp_term t =
 Term.unlabel_dp t >>= fun t -> (*to prevent two labels (head variables)*)
 Term.label_dp t
;;

let curried_term app t = 
 curried_term app t >>= fun t' ->
 match Term.root t with
  | None -> M.return t'
  | Some r -> M.is_dp r >>= fun b -> if b then sharp_term t' else M.return t'
;;
 

let curried_rule app rule = 
 let (l,r) = Rule.to_terms rule in
 curried_term app l >>= fun l ->
 curried_term app r >>= fun r ->
 M.return (Rule.of_terms l r)
;;

let curried_system app trs = 
 M.lift Trs.of_list (M.sequence (Trs.map (curried_rule app) trs))
;;

(* applicative arity for [f] *)
let aa_f trs f = aa_f f trs;;

(* applicative arity for [t] *)
let rec aa_t aa = function
 | Term.Var _ -> failwith "non left-head-variable-free term"
 | Term.Fun (f,[]) -> aa f  
 | Term.Fun (f,[t;_]) -> aa_t aa t - 1
 | _ -> failwith "term not in applicative form"
;;

(* eta saturation process *)
let eta_sat_rule trs aa rule = 
 let (l,r) = Rule.to_terms rule in
 let aa_t = aa_t aa l in
 let rec add f l r aa_t acc =
  if aa_t <= 0 then acc
  else
   fresh_var >>= fun x -> acc >>= fun acc ->
   let l' = Term.Fun (f,[l;x]) and r' = Term.Fun (f,[r;x]) in
   add f l' r' (aa_t - 1) (M.return (Rule.of_terms l' r' :: acc))
 in
 app trs >>= fun f -> add f l r aa_t (M.return [rule])
;;

let eta_sat aa trs =
 M.lift Trs.of_list (M.flat_map (eta_sat_rule trs aa) (Trs.to_list trs))
;;

let rec uncurrying_rules_f app var_list aa f =
 if aa < 0 then
  M.return Trs.empty
 else (
  let vars = List.take (aa + 1) var_list in
  assert (List.length vars > 0);
  let y, var = List.hd vars, List.tl vars in
  let i = List.length var in
  uncurry app f i >>= fun fi ->
  uncurry app f (i+1) >>= fun fi' ->
  let l = Term.Fun (app, [Term.Fun (fi, var); y]) in
  let r = Term.Fun (fi', var @ [y]) in
  (uncurrying_rules_f app var_list (aa - 1)) f >>= fun aux ->
  M.return (Trs.add (Rule.of_terms l r) aux))
;;

let uncurrying_system app var_list aa cs =
 List.foldr
  (fun f -> M.lift2 Trs.union (uncurrying_rules_f app var_list (aa f - 1) f))
  (M.return Trs.empty) cs
;;

let gen_var aa fs =
 let max_ari = List.foldr Pervasives.max 0 (List.map aa fs) in
 (M.lift List.rev) (M.sequence (List.gen (const fresh_var) max_ari))
;;

let transform aa atrs = 
 if Trs.is_empty atrs then M.return None else
 Trs.is_applicative atrs >>= fun c ->
 if c then 
  if left_head_variable_free atrs then (
   app atrs >>= fun app ->
   M.filter (M.is_ari 0) (Trs.funs atrs) >>= fun cs ->
   gen_var aa cs >>= fun vars ->

   eta_sat aa atrs >>= uncurried_system aa [app] cs >>= fun trs ->
   uncurrying_system app vars aa cs >>= 
   (M.return <.> Option.some <.> Trs.union trs)
  ) else M.return None
 else M.return None
;;

let problem_map f p = 
 let s,w = P.get_sw p in
 P.set_sw (f s) (f w) p
;;

let t_a apps atrs = 
 proper atrs >>= fun b ->
 if b then
  if Trs.is_empty atrs then M.return None else
   (* Trs.is_applicative atrs >>= fun c -> *)
   if (*c &&*) left_head_variable_free atrs then (
     let aa = aa_f atrs in (*app atrs >>= fun app ->*)
     M.filter (M.is_ari 0) (Trs.funs atrs) >>= fun cs ->
     uncurried_system aa apps cs atrs >>= fun trs ->
     M.return (Some trs)
    ) else M.return None
 else
  M.return None

let transformation_a atrs = 
  Trs.is_applicative atrs >>= fun b ->
  if b && head_variable_free atrs then 
   app atrs >>= fun app ->
   t_a [app] atrs 
  else M.return None
;;

let transformation_a_dp p =
 let s,w = P.get_sw p in
 if head_variable_free s && head_variable_free w then
  app w >>= fun appw ->
  app_dp s >>= fun apps ->
  t_a [appw] w >>= fun w' ->
  t_a apps s >>= fun s' ->
  (match s',w' with 
   | None, _ | _, None -> M.return None
   | Some s, Some w -> M.return (Some (P.set_sw s w p))
  )
 else M.return None
;;

let u1 apps p = 
 let (s,w) = P.get_sw p in
  let sw = Trs.union s w in
  (*TODO: which fs is correct DONE*)
  (*let fs = (Trs.diff (Trs.funs s) (Trs.roots s))@(Trs.funs w) in*)
  (* M.filter (M.is_ari 0) (Trs.funs w) >>= fun cs -> *)
  M.filter (M.is_ari 0) (Trs.funs sw) >>= fun cs ->
  (*TODO: vs aa_f w DONE*)
  let aa = aa_f (Trs.union s w) in
  uncurried_system aa apps cs s >>= fun s' ->
  transform aa w >>= fun w' ->
  match w' with 
   | None -> M.return None
   | Some w' -> M.return (Some (Problem.set_sw s' w' p))
;;

let transform_u1 p =
 let (_,w) = P.get_sw p in
 app w >>= fun appw ->
 u1 [appw] p
;;

let rs w = function
 | [] -> M.return true
 | t::_ -> Trs.tcap t w >>= (M.return <.> not <.> Term.is_var)
;;

let rs_rule w rule = rs w (Term.args (Rule.rhs rule));;

let strongly_root_stable w s = 
 M.sequence (List.map (rs_rule w) (Trs.to_list s)) >>= fun l ->
 M.return (List.for_all ((=) true) l)
;;

let sharp_rule rule =
 let (l,r) = Rule.to_terms rule in
 sharp_term l >>= fun l -> sharp_term r >>= fun r ->
 M.return (Rule.of_terms l r)
;;

let sharp_trs trs =
 M.sequence (List.map sharp_rule (Trs.to_list trs)) >>= fun trs ->
 M.return (Trs.of_list trs)
;;

(*TODO: demand p applicative *)
let transform_u2 p = 
 let (s,w) = P.get_sw p in
  app_dp s >>= fun apps ->
  app w >>= fun appw ->
  u1 [appw] p >>= fun p' ->
  match p' with 
   | None -> M.return None
   | Some p' ->
   let (s',w') = P.get_sw p' in
   strongly_root_stable w' s' >>= fun b ->
   if b then
    u1 apps p >>= (fun p -> match p with
     | None -> M.return None
     | Some p -> let (s,w) = P.get_sw p in
      sharp_trs s >>= fun s ->
      M.return (Some (P.set_sw s w p))
     )
   else
    M.return (Some p')
;;

let standard p = 
 let trs = P.get_trs p in
 let aa = aa_f trs in
 let f = if !(flags.a) then transformation_a else transform aa in
 f trs >>= fun trs' ->
 let p' = Option.map (flip P.set_trs p) trs' in
 M.return p'
;;

let curry p = 
 let s,w = P.get_sw p in
 is_extended_applicative s >>= fun b ->
 is_extended_applicative w >>= fun c ->
 if b && c then 
  M.return None
 else
  M.fresh_fun >>= fun app ->
  M.add_ari app 2 >>= fun _ ->
  curried_system app s >>= fun s ->
  curried_system app w >>= fun w ->
  M.return (Some (P.set_sw s w p))
;;

let dp p =
 let f = 
  if !(flags.a) then transformation_a_dp 
  else if !(flags.u1) then transform_u1
  else if !(flags.u2) then transform_u2
  else if !(flags.curry) then curry
  else fun _ -> M.return None
 in
 is_dp_applicative p >>= fun b ->
 if b then f p else M.return None
;;
 
(* new end *)

let solve p =
 let p = (if !(flags.mirror) then problem_map Trs.reflect else id) p in
 if P.is_sp p then standard p 
 else if P.is_dp p then dp p 
 else M.return None
;;

let solve fs p = 
 let configurate s = F.printf "%s@\n%!" s; flags.help := true in
 (try init (); Arg.parsex code spec fs with Arg.Bad s -> configurate s);
 if !(flags.help) then (Arg.usage spec ("Options for "^code^":"); exit 0);
 solve p >>= (M.return <.> Option.map (Pair.make p))
;;

(* Destructors *)
let get_ip = fst;;
let get_op = snd;;

(* Complexity Bounds *)
let complexity c _ = C.mul c C.constant;;

(* Compare Functions *)
let equal p q =
 P.equal (get_ip p) (get_ip q) && P.equal (get_op p) (get_op q)
;;

(* Printers *)
let fprintf fs fmt p =
 F.fprintf fmt "@[<1>%s:@\n" name; P.fprintfm fmt (get_op p) >>= fun _ ->
 F.fprintf fmt "@\n"; List.hd fs fmt >>= fun _ -> M.return (F.fprintf fmt "@]")
;;

let fprintfx fs fmt p =
 F.fprintf fmt "@{<uncurry>"; P.fprintfx fmt (get_op p) >>= fun _ ->
 List.hd fs fmt >>= fun _ -> M.return (F.fprintf fmt "@}")
;;
